/*************************************************************************/
/*************************************************************************/
/* H. Timucin OZDEMIR                                                    */
/* Date : April, 1997                                                    */
/* File Name : JSSPSolution.cc                                           */
/*************************************************************************/
/*************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <iostream.h>
#include "JSSPSolution.h"
#include "FacilitySchedule.h"
#include "deffunc.h"

JSSPSolution::~JSSPSolution()
{  myProblem = NULL;
   cout << " de allocate JSSPSolution " << endl;
   if((FacilityIndex>0)&&(FacilityScheduleList!=NULL)) 
	   delete [] FacilityScheduleList;
   FacilityIndex = 0;
   debug = 0;
   valid = 0;
}// end of constructor

JSSPSolution::JSSPSolution()
{  myProblem = NULL;
   FacilityScheduleList = NULL;
   FacilityIndex = 0;
   debug = 0;
   valid = 0;
}// end of constructor

JSSPSolution::JSSPSolution(const JSSPSolution& RHS)
{  myProblem = RHS.myProblem;
   FacilityIndex = RHS.FacilityIndex ;
   debug = RHS.debug;
   FacilityScheduleList = new FacilitySchedule[FacilityIndex];
   for(int i=0;i<FacilityIndex;i++) 
        FacilityScheduleList[i] = RHS.FacilityScheduleList[i];
   valid = 0;
}// end of constructor

void JSSPSolution::debugOn() { debug = 1; }

void JSSPSolution::debugOff() { debug = 0; }

JSSPSolution::getMaxFacilityNumber()
{  return (*myProblem).getNumberOfFacilities();
}// end of constructor()

FacilitySchedule * JSSPSolution::getFacilityScheduleIndex(int indx)
{  return (indx>=0&&indx<FacilityIndex?&FacilityScheduleList[indx]:NULL);
}// end of constructor()

void JSSPSolution::addFacilitySchedule(FacilitySchedule newSc)
{  int i;
   FacilitySchedule *Temp;
   Temp = new FacilitySchedule[FacilityIndex+1];

   valid = 0;

   if(debug) 
      cout << " I allocated the Temp FacilityIndex " << FacilityIndex << endl;

   for(i=0;i<FacilityIndex;i++) Temp[i]=FacilityScheduleList[i];

   if(debug) cout << " I copied from FacilityScheduleList to the Temp " << endl;

   Temp[i]=newSc;

   if(debug) cout << " I assigned the last one " << endl;

   if(FacilityIndex) delete [] FacilityScheduleList;
   FacilityIndex++;
   FacilityScheduleList = Temp;

   if(debug) cout << " I am out of  " << endl;

}// end of addFacilitySchedule()

JSSPSolution::JSSPSolution(JSSProblem &Problem)
{  myProblem = &Problem;
   FacilityIndex = (*myProblem).getNumberOfFacilities();
   FacilityScheduleList = new FacilitySchedule[FacilityIndex];
   valid = 0;
}// end of constructor()

void JSSPSolution::setProblem(JSSProblem &Problem)
{  myProblem = &Problem;
   FacilityIndex = (*myProblem).getNumberOfFacilities();
   FacilityScheduleList = new FacilitySchedule[FacilityIndex];
   valid = 0;
}// end of setProblem()

void JSSPSolution::setOnlyMyProblem(JSSProblem &Problem)
{  myProblem = &Problem;
   valid = 0;
}// end of setProblem()

void JSSPSolution::print()
{  int i;
   
   for(i=0;i<FacilityIndex;i++)
   {  FacilityScheduleList[i].print();
      cout << endl;
   }// end of for

}// end of print()

void JSSPSolution::addOperationToMachineIdnx(int indx, Operation newOP)
{  if(indx>=0 && indx<FacilityIndex) 
      FacilityScheduleList[indx].addOperation(newOP,0);
   else cout << "Illegal addOperation to FacilitySchedule " << endl;
   valid = 0;
}// end of addOperation()

void JSSPSolution::solve()
{  int i,j;
   int NumFac = (*myProblem).getNumberOfFacilities();
   int NumJobs = (*myProblem).getNumberOfJobs();
   int Time;

   valid = 0;
   
   for(i=0;i<NumFac;i++)
   {  
      for(j=0;j<NumJobs;j++)
      {   Operation *op = ((*myProblem).getJobList(j)).getOperation(i); 
          Operation *opPrev = ((*myProblem).getJobList(j)).getOperation(i-1); 

	  if(opPrev==NULL) Time =0;
	  else 
	    Time = FacilityScheduleList[(*opPrev).getFacilityNumber()].getTime(*opPrev);
	  FacilityScheduleList[(*op).getFacilityNumber()].addOperation(*op,Time);
      }// end of for

   }// end of for

   cout << "I solved "<< endl;

   cout << "the solution is " ;
   cout << (isFeasible()?"feasible":"not feasible") << endl;

}// end of solve

void JSSPSolution::solveRandom()
{  int i,j;
   int NumFac = (*myProblem).getNumberOfFacilities();
   int NumJobs = (*myProblem).getNumberOfJobs();
   int Time;
   int *array = new int[NumJobs];
   int first, second, tempval, rndNumberOfReplacement;
   Operation *op;
   valid = 0;


   if(debug) cout << "In the solveRandom() " << endl;

   for(i=0;i<NumFac;i++)
   {  
      if(debug) cout << "before array init " << endl;

      if(rand()%1) for(j=0;j<NumJobs;j++) array[j]=j;
      else for(j=0;j<NumJobs;j++) array[j]=NumJobs-j-1;

      if(debug) cout << "after array init Number of Jobs" << NumJobs << endl;

      rndNumberOfReplacement = rand() % (2*NumJobs) + 1;
      cout << "at Facility "<<i<<" NUmberof Replacemebt is "<<rndNumberOfReplacement << endl;
      j=0;
      while(j<rndNumberOfReplacement)
      {  
	 first  = rand() % NumJobs;
	 second = rand() % NumJobs;

	 if(debug) 
	    cout << i << " first " << first << " second " << second << endl;

	 if(first!=second) 
	 {  tempval = array[first];
	    array[first]  = array[second];
	    array[second] = tempval;
	    j++;
	 }// end of if

      }// end of while

      cout << " array : ";
         for(j=0;j<NumJobs;j++) cout << " " << array[j];
         cout << endl;

      for(j=0;j<NumJobs;j++)
      {   
	  if(debug) cout << "before getOperationWithFacilityInde " << endl;
	  op = ((*myProblem).getJobList(array[j])).getOperationWithFacilityIndex(i);
	  if(debug) cout << "after getOperationWithFacilityInde " << endl;
          if(op!=NULL)
	     FacilityScheduleList[(*op).getFacilityNumber()].addOperation(*op,0);
          else cout << "Operation is not found " << endl;
	  if(debug) cout << "after Fac : "<<i<< " Job : "<<j<< endl;

      }// end of for j 

   }// end of for i

   delete [] array;

   cout << "I randomly solved "<< endl;

   cout << "the solution is " ;
   cout << (isFeasible()?"feasible":"not feasible") << endl;

}// end of solveRandom()

JSSPSolution::isSchedulable(FacilitySchedule *newSchedule, 
			    Operation *opPtr,int *Time)
{  int ret =0;
   Operation op;
   int JobNum;
   int OpeNum;

   //cout << " in isSchedulable " << endl; 
   if(opPtr!=NULL)
   {  
      op = *opPtr;
      JobNum = op.getJobNumber();
      OpeNum = op.getOperationNumber();

      if(OpeNum==0)
      {  *Time = 0; ret=1; }
      else
       { Operation *opPrev = ((*myProblem).getJobList(JobNum)).getOperation(OpeNum-1);
         if(*Time=newSchedule[(*opPrev).getFacilityNumber()].getTime(*opPrev))
		 ret=1; 
       }// end of if

    }// end of if

   //cout << " I am out of  isSchedulable " << endl; 

   return ret;

}// end of isSchedulable()

void JSSPSolution::restore()
{
   int i, j, k;
   int NumFac = (*myProblem).getNumberOfFacilities();
   int NumJobs = (*myProblem).getNumberOfJobs();
   OperationList *OperationBag=new OperationList[NumFac]; 
   FacilitySchedule *newSchedule=new FacilitySchedule[NumFac];
   int Time, oneScheduled=0;
   valid = 0;

   cout << " I am in restore " << endl;
   // Set the maximum possible number of operations

   for(i=0;i<NumFac;i++) 
      OperationBag[i].setNumberOfOperation(NumJobs);

   for(j=0;j<NumJobs;j++)
   {  
      // place the jth job into the OperationBag of the related Facility

      for(i=0;i<NumFac;i++)
         OperationBag[i].addOperation(*(FacilityScheduleList[i].getOperationWithIndex(j)));
      
      //cout << endl << " Job# " << j << " ";

      // Now try to schedule the operations from each facility's
      // OperationBag if it is possible

      i=0;
      while(i<NumFac)
      {  
         //cout << " Fac# " << i << " ";
      
	 // So far no one is scheduled to any facility

         oneScheduled=k=0;

         while(k<OperationBag[i].getNumberOfOperation())
	 {  

	    // take the kth operation from the waiting queue

	    Operation *cur = OperationBag[i].getOperation(k); 

	    // If it is schedulable now, place it
     
	    if(isSchedulable(newSchedule,cur,&Time))
	    {  
	       //cout << " addOperation(*) " << endl;

	       newSchedule[i].addOperation(*cur,Time);

	       //cout << " after addOperation(*) " << endl;

	       OperationBag[i].deleteOperation(k); 

	       //cout << " after deleteOperation(*) " << endl;

	       //OperationBag[i].shrink(); 

	       //cout << " after shrink(*) " << endl;

	       k=0; oneScheduled = 1;

	       //cout << " after addOperation(*) " << endl;

	    } else k++;

	 }// end of while k

	 if(oneScheduled) i=0;
	 else i++;

      }// end of while i

      //cout << " end of while i " << endl;

   }// end of for

   //cout << " end of for " << endl;


   if(debug)
   for(i=0;i<NumFac;i++) 
       cout << "newSchedule ["<<i<<"] : " << newSchedule[i] << endl;


   //cout << " before delete [] FacilityScheduleList " << endl;

   if(FacilityIndex) delete [] FacilityScheduleList;

   /*FacilityScheduleList = new FacilitySchedule[NumFac];

   for(i=0;i<NumFac;i++) FacilityScheduleList[i] = newSchedule[i];
   */

   FacilityScheduleList = newSchedule;

   cout << "I solved "<< endl;

   cout << "the solution is " ;
   cout << (isFeasible()?"feasible":"not feasible") << endl;

}// end of restore()

JSSPSolution::isFeasible()
{
   int j, i, feasible=1, difference; 
   OperationList JobList;
   Operation * current ;
   Operation * previous;
   int startTimeCur;
   int startTimePre;

   if(debug) cout << "I am in isFea "<< endl;

   for(i=0; feasible && i<FacilityIndex;i++)
   {  int NumOfOper = FacilityScheduleList[i].getNumberOfOperations(); 

      if(NumOfOper)
      {  
	 previous = FacilityScheduleList[i].getOperationWithIndex(0);
	 startTimePre = FacilityScheduleList[i].getStartTime(*previous);

	 for(j=0;feasible && j<NumOfOper ;j++)
	 {  
	    current = FacilityScheduleList[i].getOperationWithIndex(j);
	    startTimeCur = FacilityScheduleList[i].getStartTime(*current);

	    if((startTimeCur-startTimePre)<(*previous).getElapsedTime())
			    feasible = 0;
            else 
              {  startTimePre = startTimeCur;
	         previous = current;
	      }  

	 }// end of for
      }// end of if

   }// end of for
	     
   for(i=0;i<(*myProblem).getNumberOfJobs() && feasible;i++)
   {
       JobList = (*myProblem).getJobList(i);
       previous = JobList.getOperation(0);
       startTimePre = 
	  FacilityScheduleList[(*previous).getFacilityNumber()].getStartTime(*previous); 

       //cout << " on machine " << i << endl;
       //cout << " s0 : " << startTimePre ;
       for(j=1;j<JobList.getNumberOfOperation() && feasible;j++)
       {
          current = JobList.getOperation(j); 
          startTimeCur = 
	  FacilityScheduleList[(*current).getFacilityNumber()].getStartTime(*current); 
						       
	  difference = startTimeCur-startTimePre;

          //cout << " s"<<j << " : " << startTimeCur ;
	  if(difference<(*previous).getElapsedTime()) feasible = 0;
	  else 
	    {  startTimePre = startTimeCur;
	       previous = current;	 
	    }	

       }// end of for

   }// end of for

   return feasible;

}// end of isFeasible()

/*
//  Locally improves the schedule.
//  Picks the machine that dictates the makespan.
//    Tries to swap the two consequtive operations and looks whether it
//    is Ok or not. If it is feasible, then calls 'restore' operation.
//
*/

void JSSPSolution::LocalImprove()
{  int maxIndx, maxT;
   int i, j;
   int NumOfOperation, first, second, notExit=1;
   Operation *preFol;
   Operation *pre;
   Operation *prePre;
   Operation *cur;
   Operation *curPre;
   int EndTimeCurPre, newStartTimeSec,  newEndTimeSec; 
   int StartTimePreFol, EndTimePrePre, newStaTimePre, newEndTimePre;


   if(FacilityIndex)
   {  maxIndx = 0; 
      maxT = FacilityScheduleList[0].getFirstEmptySlot();
      
      //* finds the facility who dominates the makespan

      for(j=1;j<FacilityIndex;j++)
	 if(FacilityScheduleList[j].getFirstEmptySlot()>maxT)
	 {  maxIndx = j; maxT = FacilityScheduleList[j].getFirstEmptySlot(); }

      //* Now apply swap operation sequentially

      NumOfOperation = FacilityScheduleList[maxIndx].getNumberOfOperations();

      first = 0;

      for(second=1;notExit && second<NumOfOperation;second++)
      {  
	 pre = FacilityScheduleList[maxIndx].getOperationWithIndex(first);
         cur = FacilityScheduleList[maxIndx].getOperationWithIndex(second);

	 curPre = (*myProblem).getPreviousOperation((*cur).getJobNumber(),*cur);
	 preFol = (*myProblem).getFollowingOperation((*pre).getJobNumber(),*pre);
	 prePre = (*myProblem).getPreviousOperation((*pre).getJobNumber(),*pre);

	 if( curPre!= NULL && preFol!= NULL )
	 {  
	    EndTimeCurPre = 
	      FacilityScheduleList[(*curPre).getFacilityNumber()].getEndTime(*curPre);
	    if(first)
	    {  
	      newStartTimeSec = maximum(EndTimeCurPre,FacilityScheduleList[maxIndx].getEndTimeOpWithIndx(first-1))+1; 
	    } else newStartTimeSec = EndTimeCurPre + 1;

	    newEndTimeSec = newStartTimeSec + (*cur).getElapsedTime();

	    StartTimePreFol = 
	      FacilityScheduleList[(*preFol).getFacilityNumber()].getStartTime(*preFol);
	    EndTimePrePre = 
	      FacilityScheduleList[(*prePre).getFacilityNumber()].getEndTime(*prePre);

	    newStaTimePre = maximum(newEndTimeSec,EndTimePrePre)+1;

	    newEndTimePre = newStaTimePre + (*pre).getElapsedTime();

	    if(newEndTimePre<StartTimePreFol && 
	       (second>(NumOfOperation-2)||newEndTimePre<FacilityScheduleList[maxIndx].getStartTimeOpWithIndx(second+1)))
	    {  FacilityScheduleList[maxIndx].swapJobs(first,second);
	       cout << first << " swapped with " << second ;
	       cout << " on machine " << maxIndx << endl;
	       restore();
	       notExit = 0;
	    }

	 } else first = second; 


      }// end of for
     
   }// end of if

}// end of LocalImprove()

JSSProblem * JSSPSolution::getMyProblem()
{  return myProblem;
}// end of getMyProblem()

JSSPSolution::getMakespan()
{
   int i;
   int maxT = 0;
   if(valid) maxT=cost;
   else
   { for(i=0;i<FacilityIndex;i++)
       maxT = maximum(maxT,FacilityScheduleList[i].getFirstEmptySlot());
     cost = maxT;
     valid = 1;
   }
   return maxT;
}// end of getMakespan()

ostream& operator<<(ostream &os, const JSSPSolution& Sol)
{  
   int i;
   int maxT = 0;
   os << " There are " << (*(Sol.myProblem)).getNumberOfFacilities() ;
   os << " number of facilities " << endl;
   for(i=0;i<(*(Sol.myProblem)).getNumberOfFacilities();i++)
   {   os << "Facility " << i << ": " << Sol.FacilityScheduleList[i] << endl;
       maxT = maximum(maxT,Sol.FacilityScheduleList[i].getFirstEmptySlot());
   }// end of for

   os << endl << " This operation takes at least " ;
   os << maxT << " time unit." << endl;

   return os;

}// end of operator<<

JSSPSolution JSSPSolution::operator =(const JSSPSolution &Sol)
{
   myProblem = Sol.myProblem;
   FacilityIndex = Sol.FacilityIndex;

   delete [] FacilityScheduleList;

   FacilityScheduleList = new FacilitySchedule[FacilityIndex];
 
   for(int i=0;i<FacilityIndex;i++)
       FacilityScheduleList[i] = Sol.FacilityScheduleList[i];
   valid = 0;

   return (*this);

}// end of operator =


JSSPSolution::operator ==(const JSSPSolution &Sol)
{  int ret = 0;

   ret = (myProblem == Sol.myProblem &&  FacilityIndex == Sol.FacilityIndex);

   for(int i=0;i<FacilityIndex && ret;i++)
       ret = (FacilityScheduleList[i] == Sol.FacilityScheduleList[i]);

   return ret;

}// end of operator ==

JSSPSolution::getNumOfJobs()
{  return (*myProblem).getNumberOfJobs();
}// end of getNumOfJobs()

JSSPSolution::getNumOfFacs()
{  return (*myProblem).getNumberOfFacilities();
}// end of getNumOfJobs()


void JSSPSolution::swapJobsOnMachine(int MacIndx,int fJob, int sJob)
{  FacilityScheduleList[MacIndx].swapJobs(fJob,sJob);
   valid = 0;
}

void JSSPSolution::swapJobsOrderOnMachine(int MacIndx,int fJob, int sJob)
{  FacilityScheduleList[MacIndx].swapJobsOrder(fJob,sJob);
   valid = 0;
}


